from sage.all import *
from itertools import islice, count
from libc.stdlib cimport free, malloc
from legendre_symbol import legendre_symbol

# Zu Anfang wird die Grenze B des Quadratischen Siebs berechnet,
# welche die Groesse der Faktorbasis und die Laenge M = B*B des
# Siebintervalls I bestimmt.
def calculate_boundary(number):

    return ceil(sqrt(e**(sqrt(ln(number)*ln(ln(number))))))

# Anschliessend wird die Faktorbasis erzeugt, welcher die Zahlen
# -1 und 2 stets angehoeren. Mit Hilfe des Legendre-Symbols werden
# zudem nur jene Primzahlen in die Faktorbasis aufgenommen, die
# im spaeteren Siebvorgang auch tatsaechlich als Teiler einer Zahl
# q(x) = x*x - number auftreten koennen.
def generate_factor_base(number, boundary):

    factor_base = [-1, 2]

    for prime in prime_range(3, boundary):
        if legendre_symbol(number, prime) == 1:
            factor_base.append(prime)

        # Statt des Legendre-Symbols kann auch das Euler-
        # Kriterium - number**((prime - 1)//2)%prime -
        # herangezogen werden. Die Berechnung wuerde dann
        # jedoch wesentlich laenger dauern.

    return factor_base

# Das Siebintervall I wird nun hinsichtlich B-glatter Zahlen
# q(x) gesiebt. Bei Bedarf kann das Siebintervall I in Teil-
# intervalle zerlegt werden, um den Speicherbedarf - bei ueber-
# schaubarem Leistungsverlust - zu reduzieren.
def sieve(number, boundary, factor_base, int number_of_intervals = 1):

    start_point = ceil(sqrt(number)) - (boundary**2)//2
    end_point = start_point + 2*(boundary**2)//2
    # Der Schwellwert legt fest, ab welchem Wert eine Zahl q(x),
    # deren ungefaehrer 2er-Logarithmus im Siebvorgang bestimmt
    # wird, als B-glatt vermutet und mit Probedivision ueberprueft
    # werden soll.
    threshold = ceil(log((start_point + ((end_point - start_point)//5)*3)**2 - number, 2)) - 30

    sieved_array = []
    interval_length = (end_point - start_point)//number_of_intervals

    end_point = start_point - 1
    for interval in islice(count(0), number_of_intervals):
        start_point = end_point + 1
        end_point = start_point + interval_length
        sieved_array.append(__sieve_interval(number, factor_base, start_point, end_point, threshold))

    # Flachklopfen des aus - moeglicherweise mehreren - Teil-
    # intervallen des Siebintervalls I bestehenden Arrays.
    return reduce(lambda x, y: x + y, sieved_array)

def __sieve_interval(number, factor_base, start_point, end_point, threshold):

    # Zur Reduzierung des Speicherbedarfs wird festgelegt,
    # dass der ungefaehre 2er-Logarithmus einer Zahl q(x)
    # max. 2 Byte verbrauchen darf.
    cdef short logarithm

    sieving_array_length = end_point - start_point + 1
    logarithms = <unsigned short *> malloc(sizeof(unsigned short)*sieving_array_length)

    # Um die Geschwindigkeit deutlich zu steigern, werden die
    # kleinsten Primzahlen der Faktorbasis vernachlaessigt. Zum
    # Ausgleich wurde der zuvor festgelegte Schwellwert reduziert.
    for prime in factor_base[9:]:
        # Bestimmen der Loesungen +x und -x der quadratischen
        # Kongruenz x^2 = number (mod prime), um zu erfahren,
        # welche Zahlen q(x) ein Vielfaches von prime sind.
        square_roots = mod(number, prime).sqrt(all = True)

        found = 0
        logarithm = round(log(prime, 2))
        # Statt den 2er-Logarithmus jeder einzelnen Zahl q(x) zu
        # berechnen, werden nur die ersten 2 Zahlen q(x) gesucht,
        # die ein Vielfaches von prime sind, denn dann sind auch
        # alle Zahlen q(x + k*prime), k ist eine natuerliche Zahl
        # groesser 0, ein Vielfaches von prime.
        for index in islice(count(0), prime):

            if (start_point + index)%prime == square_roots[0]:
                for multiple in islice(count(index, prime), (sieving_array_length - index + prime - 1)//prime):
                    logarithms[multiple] += logarithm
                found += 1

            if (start_point + index)%prime == square_roots[1]:
                for multiple in islice(count(index, prime), (sieving_array_length - index + prime - 1)//prime):
                    logarithms[multiple] += logarithm
                found += 1

            if found == 2: break

    sieved_array = []
    for index in islice(count(0), sieving_array_length):
        # Anhand des Schwellwerts wird nun entschieden, welche
        # Zahlen q(x) vermutlich B-glatt sind. Diese Vermutung
        # wird anschliessend mittels Probedivision ueberprueft
        # und die Faktorisierung bei Erfolg zurueckgegeben.
        if logarithms[index] >= threshold:
            composite = (start_point + index)**2 - number
            factorization = []
            if composite < 0:
                composite *= -1
                factorization.append(-1)
            for prime in factor_base[1:]:
                while composite%prime == 0:
                    composite /= prime
                    factorization.append(prime)
            if composite == 1:
                sieved_array.append([start_point + index, factorization])

    free(logarithms)

    return sieved_array

# Zu jeder B-glatten Zahl wird ein Exponentenvektor angelegt,
# der angibt, wie oft eine Primzahl der Faktorbasis in einer
# Zahl q(x) enthalten ist.
def generate_exponent_vectors(factor_base, sieved_array):

    exponent_vectors = [[0 for index in islice(count(0), len(factor_base))] for index in islice(count(0), len(sieved_array))]

    for index in islice(count(0), len(sieved_array)):
        for factor in sieved_array[index][1]:
            exponent_vectors[index][factor_base.index(factor)] += 1

    return exponent_vectors

# Die Exponentenvektoren der B-glatten Zahlen werden mod 2
# reduziert und der Kernel der aus ihnen gebildeten Matrix
# bestimmt, also all jene Vektoren, die, multipliziert mit
# der Matrix, den Nullvektor ergeben. Anders ausgedrueckt:
# Es werden die Kombinationen von Zahlen q(x) gesucht, die
# miteinander multipliziert eine Quadratzahl sind.
def generate_matrix(exponent_vectors):

    return matrix(GF(2), exponent_vectors).kernel().basis()

# Zum Schluss wird die Zahl number in zwei Faktoren zerlegt,
# die, sollten sie keine Primzahlen sein, wiederum in das
# Quadratische Sieb oder ein anderes Faktorisierungsver-
# fahren eingegeben werden koennen.
def factorize(number, sieved_array, matrix):

    for row in matrix:
        x = []
        y = []
        for index in islice(count(0), len(row)):
            if row[index] == 1:
                x.append(sieved_array[index][0])
                y.append(sieved_array[index][1])
        # Da die Faktorisierung der miteinander multiplizierten
        # Zahlen q(x) bekannt ist und jeder Faktor geradzahlig oft
        # vorkommt, kann die Quadratwurzel effizient bestimmt werden:
        # jeder zweite Faktor wird verworfen und die verbliebenen
        # Faktoren miteinander multipliziert.
        y = sorted(reduce(lambda x, y: x + y, y))[::2]
        y = reduce(lambda i, j: i*j, y)
        x = reduce(lambda i, j: i*j, x)
        factor = gcd((x%number) - (y%number), number)
        if (factor != 1) and (factor != number):
            return [factor, number/factor]

    return []
